#include "common.h"
#include "util.h"
#include "parser.h"
#include "dump_ast.h"

#include <boost/spirit/iterator/position_iterator.hpp>
#include <boost/spirit/utility/confix.hpp>
#include <boost/spirit/utility/escape_char.hpp>

using namespace boost::spirit;

struct GrammarImpl : public grammar<GrammarImpl>, Grammar {
    template <class Scanner>
    struct definition {
#define DEF_RULE(name) \
typedef rule<Scanner, parser_context<>, parser_tag<name##ID> > name##_rule_t; \
name##_rule_t name;
#include "rules.tab"
#undef DEF_RULE

        // why doesn't inner_node_d work for some cases?
#define INNER_NODE_D(a, b, c) \
(discard_node_d[ch_p(a)] >> (b) >> discard_node_d[ch_p(c)])
#define INNER_PAREN_D(b) INNER_NODE_D('(', b, ')')

        definition(const Grammar& self) {
            ident =
                token_node_d[
                    lexeme_d[
                        (alpha_p | ch_p('_')) >> *(alnum_p | ch_p('_'))
                        ]];
            assign = ident >> discard_node_d[ch_p('=')] >> expr;
            int_expr = int_p;
            str_expr =
                discard_node_d[str_p("\"\"")] >> str_p("") |
                token_node_d[
                    lexeme_d[
                        inner_node_d[
                            confix_p('"', *c_escape_ch_p, '"')
                            ]]];
            funcall = ident >> (
                ('(' >> discard_node_d[ch_p(')')]) |  // ugly
                INNER_PAREN_D(expr % discard_node_d[ch_p(',')]));
            fact =
                INNER_PAREN_D(expr) |
                assign |
                funcall |
                int_expr |
                str_expr |
                ident;
            md_expr = fact % (ch_p('*')|ch_p('/')|ch_p('%'));
            pm_expr = md_expr % (ch_p('+')|ch_p('-'));
            cmp_expr = pm_expr % (str_p("==")|"<="|">="|"!="|'>'|'<');
            cand_expr = cmp_expr;
            cor_expr = cand_expr;
            expr = cmp_expr;
            scope_stmt =
                INNER_NODE_D('{', *stmt, '}');
            else_stmt =
                discard_node_d[str_p("else")] >> scope_stmt;
            if_stmt =
                discard_node_d[str_p("if")] >>
                INNER_PAREN_D(expr) >>
                scope_stmt >>
                ((discard_node_d[*ch_p('\n')] >> else_stmt) | eps_p);
            while_stmt =
                discard_node_d[str_p("while")] >>
                INNER_PAREN_D(expr) >>
                scope_stmt;
            stmt =
                discard_node_d[*ch_p('\n')] >> (
                    if_stmt |
                    while_stmt |
                    expr
                    ) >> discard_node_d[+ch_p('\n')];
            program = *stmt;
        }

        const program_rule_t& start() const { return program; }
    };
};

parser_info parse_program(const char* program) {
    GrammarImpl parser;
    typedef position_iterator<char const*> iterator_t;
    iterator_t begin(program, program + strlen(program));
    iterator_t end;
    parser_info info =
        ast_parse(begin, end, parser, ch_p(' ')|ch_p('\t'));

#ifndef NDEBUG
    dump_ast(info);
#endif

    if (!info.full) {
        error((format("syntax error @%d:%d")
               % info.stop.get_position().line
               % info.stop.get_position().column).str());
    }
    return info;
}
